#pragma once

#include "../net/stream.hpp"

#include <boost/asio.hpp>

#include <openssl/evp.h>
#include <openssl/x509.h>

#include <functional>
#include <memory>
#include <optional>
#include <vector>

struct lua_State;

template <typename T, int(*UpRef)(T*), void(*Free)(T*)>
class Ref {
    struct Deleter { auto operator()(auto ptr) { Free(ptr); }};
    std::unique_ptr<T, Deleter> obj;
public:
    Ref() = default;
    Ref(T* t) : obj{t} { if (t) UpRef(t); }
    auto get() const -> T* { return obj.get(); }
};

struct Settings
{
    bool tls;
    std::string host;
    std::uint16_t port;

    Ref<X509, X509_up_ref, X509_free> client_cert;
    Ref<EVP_PKEY, EVP_PKEY_up_ref, EVP_PKEY_free> client_key;
    std::string verify;
    std::string sni;

    std::string socks_host;
    std::uint16_t socks_port;
    std::string socks_user;
    std::string socks_pass;
};

class irc_connection final : public std::enable_shared_from_this<irc_connection>
{
public:
    using stream_type = Stream;
    static std::size_t const irc_buffer_size = 131'072;

private:
    stream_type stream_;
    boost::asio::ip::tcp::resolver resolver_;
    std::vector<int> write_refs;
    std::vector<boost::asio::const_buffer> write_buffers;
    lua_State* L;
    bool writing_;

    struct Private
    {
    };

public:
    irc_connection(Private, boost::asio::io_context&, lua_State*);
    ~irc_connection();

    auto operator=(irc_connection const&) -> irc_connection& = delete;
    auto operator=(irc_connection&&) -> irc_connection& = delete;
    irc_connection(irc_connection const&) = delete;
    irc_connection(irc_connection&&) = delete;

    auto static create(boost::asio::io_context& io_context, lua_State* const L) -> std::shared_ptr<irc_connection>
    {
        return std::make_shared<irc_connection>(Private{}, io_context, L);
    }

    auto get_stream() -> stream_type&
    {
        return stream_;
    }

    auto get_lua() const -> lua_State*
    {
        return L;
    }

    // Queue messages for writing
    /**
     * @brief Write a message to the output stream
     *
     * @param msg The string to write including any needed line-terminators
     * @param ref A reference generated by luaL_ref keeping this string valid
     */
    auto write(std::string_view msg, int ref) -> void;

    /**
     * @brief Abruptly close the connection.
     * 
     */
    auto close() -> void;

    /**
     * @brief Initiate a connection to the IRC server
     * 
     * @param settings Parameters needed to establish a text stream with the server.
     * @return Space-separated, key=value pairs describing the connection
     */
    auto connect(Settings settings) -> boost::asio::awaitable<std::string>;

private:
    // There's data now, actually write it
    auto write_actual() -> void;
};
